/**
 * Copyright (c) 2016-present, Facebook, Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef CAFFE2_UTILS_BENCH_UTILS_H_
#define CAFFE2_UTILS_BENCH_UTILS_H_

#include <stdint.h>

#include <c10/macros/Export.h>

namespace caffe2 {

TORCH_API uint32_t wipe_cache();

} // namespace caffe2

#endif // CAFFE2_UTILS_BENCH_UTILS_H_
